/*******************************************************************************
 * Copyright (c) 2004, 2013 John Krasnay and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     John Krasnay - initial API and implementation
 *     Igor Jacy Lino Campista - Java 5 warnings fixed (bug 311325)
 *     Carsten Hiesserich - Fixed layout issue when editing nested blocks (bug 408482)
 *******************************************************************************/
package org.eclipse.vex.core.internal.layout;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.eclipse.vex.core.internal.core.Caret;
import org.eclipse.vex.core.internal.core.Color;
import org.eclipse.vex.core.internal.core.ColorResource;
import org.eclipse.vex.core.internal.core.FontMetrics;
import org.eclipse.vex.core.internal.core.Graphics;
import org.eclipse.vex.core.internal.core.Insets;
import org.eclipse.vex.core.internal.css.StyleSheet;
import org.eclipse.vex.core.internal.css.Styles;
import org.eclipse.vex.core.provisional.dom.BaseNodeVisitorWithResult;
import org.eclipse.vex.core.provisional.dom.ContentPosition;
import org.eclipse.vex.core.provisional.dom.ContentRange;
import org.eclipse.vex.core.provisional.dom.IComment;
import org.eclipse.vex.core.provisional.dom.IDocument;
import org.eclipse.vex.core.provisional.dom.IElement;
import org.eclipse.vex.core.provisional.dom.INode;
import org.eclipse.vex.core.provisional.dom.IParent;
import org.eclipse.vex.core.provisional.dom.IPosition;
import org.eclipse.vex.core.provisional.dom.IProcessingInstruction;

/**
 * Base class of block boxes that can contain other block boxes. This class implements the layout method and various
 * navigation methods. Subclasses must implement the createChildren method.
 *
 * Subclasses can be anonymous or non-anonymous (i.e. generated by an element). Since the vast majority of instances
 * will be non-anonymous, this class can manage the element and top and bottom margins without too much inefficiency.
 *
 * <p>
 * Subclasses that can be anonymous must override the getStartPosition and getEndPosition classes to return the range
 * covered by the box.
 * </p>
 */
public abstract class AbstractBlockBox extends AbstractBox implements BlockBox {

	/** The length, in pixels, of the horizontal caret between block boxes */
	private static final int H_CARET_LENGTH = 20;

	/**
	 * Element with which we are associated. For anonymous boxes, this is null.
	 */
	private final INode node;

	/*
	 * We cache the top and bottom margins, since they may be affected by our children.
	 */
	private int marginTop;
	private int marginBottom;

	/**
	 * Start position of an anonymous box. For non-anonymous boxes, this is null.
	 */
	private IPosition startPosition;

	/**
	 * End position of an anonymous box. For non-anonymous boxes, this is null.
	 */
	private IPosition endPosition;

	/**
	 * Class constructor for non-anonymous boxes.
	 *
	 * @param context
	 *            LayoutContext being used.
	 * @param parent
	 *            Parent box.
	 * @param node
	 *            Node associated with this box. anonymous box.
	 */
	public AbstractBlockBox(final LayoutContext context, final BlockBox parent, final INode node) {
		this.parent = parent;
		this.node = node;

		final Styles styles = context.getStyleSheet().getStyles(node);
		final int parentWidth = parent.getWidth();
		marginTop = styles.getMarginTop().get(parentWidth);
		marginBottom = styles.getMarginBottom().get(parentWidth);

	}

	/**
	 * Class constructor for anonymous boxes.
	 *
	 * @param context
	 *            LayoutContext to use.
	 * @param parent
	 *            Parent box.
	 * @param startOffset
	 *            Start of the range covered by the box.
	 * @param endOffset
	 *            End of the range covered by the box.
	 */
	public AbstractBlockBox(final LayoutContext context, final BlockBox parent, final int startOffset, final int endOffset) {
		this.parent = parent;
		node = null;
		marginTop = 0;
		marginBottom = 0;

		final IDocument doc = context.getDocument();
		startPosition = doc.createPosition(startOffset);
		endPosition = doc.createPosition(endOffset);
	}

	/**
	 * Walks the box tree and returns the nearest enclosing element.
	 */
	protected IParent findContainingParent() {
		INode node = getNode();
		while (!(node instanceof IParent)) {
			node = node.getParent();
		}
		return (IParent) node;
	}

	/**
	 * Returns this box's children as an array of BlockBoxes.
	 */
	protected BlockBox[] getBlockChildren() {
		return (BlockBox[]) getChildren();
	}

	@Override
	public Caret getCaret(final LayoutContext context, final ContentPosition position) {
		if (getLayoutState() != LAYOUT_OK) {
			return newHCaret(estimateCaretPosition(position));
		}

		Box lastBox = null;
		for (final Box box : getChildrenWithContent()) {
			if (isPositionBefore(position, box)) {
				return newHCaret(caretPositionBetween(lastBox, box));
			}

			if (isPositionWithin(position, box)) {
				return getChildCaret(context, position, box);
			}

			lastBox = box;
		}

		return newHCaret(caretPositionAtEnd());
	}

	private HCaret newHCaret(final int y) {
		return new HCaret(0, y, getHCaretWidth());
	}

	private int estimateCaretPosition(final ContentPosition position) {
		final int relativeOffset = position.getOffset() - getStartOffset();
		final int charCount = getCharCount();
		if (charCount > 0) {
			return getHeight() * relativeOffset / charCount;
		}
		return 0;
	}

	private static int caretPositionBetween(final Box box1, final Box box2) {
		if (box1 == null || box2 == null) {
			return 0;
		}
		return (box1.getY() + box1.getHeight() + box2.getY()) / 2;
	}

	private static Caret getChildCaret(final LayoutContext context, final ContentPosition position, final Box box) {
		final Caret caret = box.getCaret(context, position);
		caret.translate(box.getX(), box.getY());
		return caret;
	}

	private int caretPositionAtEnd() {
		final int y;
		if (hasChildren()) {
			y = getHeight();
		} else {
			y = getHeight() / 2;
		}
		return y;
	}

	private static boolean isPositionBefore(final ContentPosition position, final Box box) {
		return position.getOffset() < box.getStartOffset();
	}

	private static boolean isPositionWithin(final ContentPosition position, final Box box) {
		return position.getOffset() >= box.getStartOffset() && position.getOffset() <= box.getEndOffset();
	}

	private int getCharCount() {
		return getEndOffset() - getStartOffset();
	}

	@Override
	public Box[] getChildren() {
		return children;
	}

	protected BlockBox[] getChildrenWithContent() {
		final Box[] children = getChildren();
		final List<BlockBox> result = new ArrayList<BlockBox>(children.length);
		for (final Box child : children) {
			if (child.hasContent()) {
				result.add((BlockBox) child);
			}
		}
		return result.toArray(new BlockBox[result.size()]);
	}

	@Override
	public INode getNode() {
		if (isAnonymous()) {
			return parent.getNode();
		}

		return node;
	}

	@Override
	public int getEndOffset() {
		if (!isAnonymous()) {
			return getNode().getEndOffset();
		} else if (getEndPosition() != null) {
			return getEndPosition().getOffset();
		} else {
			throw new IllegalStateException();
		}
	}

	/**
	 * Returns the estimated size of the box, based on the the current font size and the number of characters covered by
	 * the box. This is a utility method that can be used in implementation of setInitialSize. It assumes the width of
	 * the box has already been correctly set.
	 *
	 * @param context
	 *            LayoutContext to use.
	 */
	protected int getEstimatedHeight(final LayoutContext context) {

		final INode node = findContainingParent();
		final Styles styles = context.getStyleSheet().getStyles(node);
		final int charCount = getCharCount();

		final float fontSize = styles.getFontSize();
		final float lineHeight = styles.getLineHeight();
		final float estHeight = lineHeight * fontSize * 0.6f * charCount / getWidth();

		return Math.round(Math.max(estHeight, lineHeight));
	}

	@Override
	public LineBox getFirstLine() {
		if (hasChildren()) {
			final BlockBox firstChild = (BlockBox) getChildren()[0];
			return firstChild.getFirstLine();
		} else {
			return null;
		}
	}

	/**
	 * Returns the width of the horizontal caret. This is overridden by TableBox to return a caret that is the full
	 * width of the table.
	 */
	protected int getHCaretWidth() {
		return H_CARET_LENGTH;
	}

	@Override
	public Insets getInsets(final LayoutContext context, final int containerWidth) {

		if (getNode() != null) {
			final Styles styles = context.getStyleSheet().getStyles(getNode());

			final int top = marginTop + styles.getBorderTopWidth() + styles.getPaddingTop().get(containerWidth);

			final int left = styles.getMarginLeft().get(containerWidth) + styles.getBorderLeftWidth() + styles.getPaddingLeft().get(containerWidth);

			final int bottom = marginBottom + styles.getBorderBottomWidth() + styles.getPaddingBottom().get(containerWidth);

			final int right = styles.getMarginRight().get(containerWidth) + styles.getBorderRightWidth() + styles.getPaddingRight().get(containerWidth);

			return new Insets(top, left, bottom, right);
		} else {
			return new Insets(marginTop, 0, marginBottom, 0);
		}
	}

	@Override
	public LineBox getLastLine() {
		if (hasChildren()) {
			final BlockBox lastChild = (BlockBox) getChildren()[getChildren().length - 1];
			return lastChild.getLastLine();
		} else {
			return null;
		}
	}

	/**
	 * Returns the layout state of this box.
	 */
	public byte getLayoutState() {
		return layoutState;
	}

	@Override
	public ContentPosition getLineEndPosition(final ContentPosition linePosition) {
		final BlockBox[] children = getChildrenWithContent();
		for (final BlockBox element2 : children) {
			if (element2.containsPosition(linePosition)) {
				return element2.getLineEndPosition(linePosition);
			}
		}
		return linePosition;
	}

	@Override
	public ContentPosition getLineStartPosition(final ContentPosition linePosition) {
		final BlockBox[] children = getChildrenWithContent();
		for (final BlockBox element2 : children) {
			if (element2.containsPosition(linePosition)) {
				return element2.getLineStartPosition(linePosition);
			}
		}
		return linePosition;
	}

	@Override
	public int getMarginBottom() {
		return marginBottom;
	}

	@Override
	public int getMarginTop() {
		return marginTop;
	}

	@Override
	public ContentPosition getNextLinePosition(final LayoutContext context, final ContentPosition linePosition, final int x) {

		//
		// This algorithm works when this block owns the offsets between
		// its children.
		//

		if (linePosition.getOffset() == getEndOffset()) {
			// We are at the end of this box - no line here
			return null;
		}

		final BlockBox[] children = getChildrenWithContent();

		if (linePosition.getOffset() < getStartOffset() && children.length > 0 && children[0].getStartOffset() > getStartOffset()) {
			// If there's an offset before the first child, put the caret there.
			return getNode().getStartPosition().moveBy(1);
		}

		for (final BlockBox child : children) {
			if (linePosition.getOffset() <= child.getEndOffset()) {
				final ContentPosition newPosition = child.getNextLinePosition(context, linePosition, x - child.getX());
				if (newPosition == null) {
					// No next line in child box
					// Return the start position of the next node
					return child.getNode().getEndPosition().moveBy(1);
				} else {
					return newPosition;
				}
			}
		}

		return new ContentPosition(getNode(), getEndOffset());
	}

	@Override
	public BlockBox getParent() {
		return parent;
	}

	@Override
	public ContentPosition getPreviousLinePosition(final LayoutContext context, final ContentPosition linePosition, final int x) {

		if (linePosition.getOffset() == getStartOffset()) {
			// We are at the start of this box - no previous line here
			return null;
		}

		final BlockBox[] children = getChildrenWithContent();

		if (linePosition.getOffset() > getEndOffset() && children.length > 0 && children[children.length - 1].getEndOffset() < getEndOffset()) {
			// If there's an offset after the last child, put the caret there.
			// This may be an anonymous box without an associated node - Find a parent with an associated node
			return getNode().getEndPosition();
		}

		for (int i = children.length; i > 0; i--) {
			final BlockBox child = children[i - 1];
			if (linePosition.getOffset() >= child.getStartOffset()) {
				final ContentPosition newPosition = child.getPreviousLinePosition(context, linePosition, x - child.getX());
				if (newPosition == null && i > 0) {
					// No previous line in child box
					// Return the child's start position. Insertions at this position will insert before the node.
					return child.getNode().getStartPosition();
				} else {
					return newPosition;
				}
			}
		}

		// No child returned a position, so place the caret before any child
		return new ContentPosition(getNode(), getStartOffset());
	}

	@Override
	public int getStartOffset() {
		if (!isAnonymous()) {
			return getNode().getStartOffset() + 1;
		} else if (getStartPosition() != null) {
			return getStartPosition().getOffset();
		} else {
			throw new IllegalStateException();
		}
	}

	@Override
	public boolean hasContent() {
		if (isAnonymous()) {
			return hasChildren();
		}
		return getNode().isAssociated();
	}

	@Override
	public void invalidate(final boolean direct) {

		if (direct) {
			layoutState = LAYOUT_REDO;
		} else {
			layoutState = LAYOUT_PROPAGATE;
		}

		if (getParent() instanceof AbstractBlockBox) {
			((AbstractBlockBox) getParent()).invalidate(false);
		}
	}

	@Override
	public boolean isAnonymous() {
		return node == null;
	}

	/**
	 * Call the given callback for each child matching one of the given display styles. Any nodes that do not match one
	 * of the given display types cause the onRange callback to be called, with a range covering all such contiguous
	 * nodes.
	 *
	 * @param styleSheet
	 *            StyleSheet from which to determine display styles.
	 * @param displayStyles
	 *            Display types to be explicitly recognized.
	 * @param callback
	 *            DisplayStyleCallback through which the caller is notified of matching elements and non-matching
	 *            ranges.
	 */
	protected void iterateChildrenByDisplayStyle(final StyleSheet styleSheet, final Set<String> displayStyles, final ElementOrRangeCallback callback) {
		LayoutUtils.iterateChildrenByDisplayStyle(styleSheet, displayStyles, findContainingParent(), getStartOffset(), getEndOffset(), callback);
	}

	@Override
	public void paint(final LayoutContext context, final int x, final int y) {

		if (skipPaint(context, x, y)) {
			return;
		}

		final boolean drawBorders = !context.isNodeSelected(getNode());

		this.drawBox(context, x, y, getParent().getWidth(), drawBorders);

		paintChildren(context, x, y);

		paintSelectionFrame(context, x, y, true);
	}

	/**
	 * Default implementation. Width is calculated as the parent's width minus this box's insets. Height is calculated
	 * by getEstimatedHeight.
	 */
	@Override
	public void setInitialSize(final LayoutContext context) {
		final int parentWidth = getParent().getWidth();
		final Insets insets = this.getInsets(context, parentWidth);
		setWidth(Math.max(0, parentWidth - insets.getLeft() - insets.getRight()));
		setHeight(getEstimatedHeight(context));
	}

	@Override
	public ContentPosition viewToModel(final LayoutContext context, final int x, final int y) {
		final Box[] children = getChildren();

		if (children == null) {
			final int charCount = getCharCount() - 1;
			if (charCount == 0 || getHeight() == 0) {
				return getNode().getEndPosition();
			}

			return new ContentPosition(getNode(), getStartOffset() + charCount * y / getHeight());
		}

		for (final Box child : children) {
			if (!child.hasContent()) {
				continue;
			}

			if (y < child.getY()) {
				return child.getNode().getStartPosition().moveBy(-1);
			}

			if (y < child.getY() + child.getHeight()) {
				return child.viewToModel(context, x - child.getX(), y - child.getY());
			}

		}

		return getNode().getEndPosition().copy();
	}

	// ===================================================== PRIVATE

	private final BlockBox parent;
	private Box[] children;

	/**
	 * Paint a frame that indicates a block element box has been selected.
	 *
	 * @param context
	 *            LayoutContext to use.
	 * @param x
	 *            x-coordinate at which to draw
	 * @param y
	 *            y-coordinate at which to draw.
	 * @param selected
	 */
	protected void paintSelectionFrame(final LayoutContext context, final int x, final int y, final boolean selected) {

		final INode node = getNode();
		final IParent parent = node == null ? null : node.getParent();

		final boolean paintFrame = context.isNodeSelected(node) && !context.isNodeSelected(parent);

		if (!paintFrame) {
			return;
		}

		final Graphics g = context.getGraphics();
		ColorResource foreground;
		ColorResource background;

		if (selected) {
			foreground = g.getSystemColor(ColorResource.SELECTION_FOREGROUND);
			background = g.getSystemColor(ColorResource.SELECTION_BACKGROUND);
		} else {
			foreground = g.createColor(new Color(0, 0, 0));
			background = g.createColor(new Color(0xcc, 0xcc, 0xcc));
		}

		final FontMetrics fm = g.getFontMetrics();
		final ColorResource oldColor = g.setColor(background);
		g.setLineStyle(Graphics.LINE_SOLID);
		g.setLineWidth(1);
		final String frameName = getSelectionFrameName(node);
		final int tabWidth = g.stringWidth(frameName) + fm.getLeading();
		final int tabHeight = fm.getHeight();
		final int tabX = x + getWidth() - tabWidth;
		final int tabY = y + getHeight() - tabHeight;
		g.drawRect(x, y, getWidth(), getHeight());
		g.fillRect(tabX, tabY, tabWidth, tabHeight);
		g.setColor(foreground);
		g.drawString(frameName, tabX + fm.getLeading() / 2, tabY);

		g.setColor(oldColor);
		if (!selected) {
			foreground.dispose();
			background.dispose();
		}
	}

	protected String getSelectionFrameName(final INode node) {
		return node.accept(new BaseNodeVisitorWithResult<String>() {
			@Override
			public String visit(final IElement element) {
				return element.getPrefixedName();
			}

			@Override
			public String visit(final IComment comment) {
				return "Comment";
			}

			@Override
			public String visit(final IProcessingInstruction comment) {
				return "Processing instruction";
			}
		});
	}

	/** Layout is OK */
	public static final byte LAYOUT_OK = 0;

	/** My layout is OK, but one of my children needs to be laid out */
	public static final byte LAYOUT_PROPAGATE = 1;

	/** I need to be laid out */
	public static final byte LAYOUT_REDO = 2;

	private byte layoutState = LAYOUT_REDO;

	@Override
	public VerticalRange layout(final LayoutContext context, final int top, final int bottom) {

		VerticalRange repaintRange = null;
		boolean repaintToBottom = false;
		final int originalHeight = getHeight();

		if (layoutState == LAYOUT_REDO) {

			final List<Box> childList = createChildren(context);
			children = childList.toArray(new BlockBox[childList.size()]);

			// Even though we position children after layout, we have to
			// do a preliminary positioning here so we now which ones
			// overlap our layout band
			for (final Box element2 : children) {
				final BlockBox child = (BlockBox) element2;
				child.setInitialSize(context);
			}
			positionChildren(context);

			// repaint everything
			repaintToBottom = true;
			repaintRange = new VerticalRange(0, 0);
		}

		final Box[] children = getChildren();
		for (final Box element2 : children) {
			if (element2 instanceof BlockBox) {
				final BlockBox child = (BlockBox) element2;
				if (top <= child.getY() + child.getHeight() && bottom >= child.getY()) {

					final VerticalRange layoutRange = child.layout(context, top - child.getY(), bottom - child.getY());
					if (layoutRange != null) {
						if (repaintRange == null) {
							repaintRange = layoutRange.moveBy(child.getY());
						} else {
							repaintRange = repaintRange.union(layoutRange.moveBy(child.getY()));
						}
					}
				}
			}
		}

		final int childRepaintStart = positionChildren(context);
		if (childRepaintStart != -1) {
			repaintToBottom = true;
			repaintRange = new VerticalRange(Math.min(repaintRange.getTop(), childRepaintStart), repaintRange.getBottom());
		}

		layoutState = LAYOUT_OK;

		if (repaintToBottom) {
			repaintRange = new VerticalRange(repaintRange.getTop(), Math.max(originalHeight, getHeight()));
		}

		if (repaintRange == null || repaintRange.isEmpty()) {
			return null;
		}

		return repaintRange;
	}

	protected abstract List<Box> createChildren(LayoutContext context);

	/**
	 * Creates a list of block boxes for a given document range. beforeInlines and afterInlines are prepended/appended
	 * to the first/last block child, and each may be null.
	 */
	protected List<Box> createBlockBoxes(final LayoutContext context, final int startOffset, final int endOffset, final int width, final List<InlineBox> beforeInlines,
			final List<InlineBox> afterInlines) {
		final List<Box> blockBoxes = new ArrayList<Box>();
		final List<InlineBox> pendingInlines = new ArrayList<InlineBox>();

		if (beforeInlines != null) {
			pendingInlines.addAll(beforeInlines);
		}

		final IDocument document = context.getDocument();
		final INode node = document.findCommonNode(startOffset, endOffset);

		if (startOffset == endOffset) {
			final int relOffset = startOffset - node.getStartOffset();
			pendingInlines.add(new PlaceholderBox(context, node, relOffset));
		} else if (node instanceof IParent) {
			final FormattingPortionIterator iter = new FormattingPortionIterator(context.getWhitespacePolicy(), (IParent) node, startOffset, endOffset);
			while (true) {
				Object formattingPortion = iter.next();
				if (formattingPortion == null) {
					break;
				}

				if (formattingPortion instanceof ContentRange) {
					final ContentRange range = (ContentRange) formattingPortion;
					final InlineElementBox.InlineBoxes inlineBoxes = InlineElementBox.createInlineBoxes(context, node, range);
					pendingInlines.addAll(inlineBoxes.boxes);
					pendingInlines.add(new PlaceholderBox(context, node, range.getEndOffset() - node.getStartOffset()));
				} else {
					if (!pendingInlines.isEmpty()) {
						blockBoxes.add(ParagraphBox.create(context, node, pendingInlines, width));
						pendingInlines.clear();
					}

					if (isTableChild(context, formattingPortion)) {
						// Consume continguous table children and create an
						// anonymous table.
						final int tableStartOffset = ((IElement) formattingPortion).getStartOffset();
						int tableEndOffset = -1; // dummy to hide warning
						while (isTableChild(context, formattingPortion)) {
							tableEndOffset = ((IElement) formattingPortion).getEndOffset() + 1;
							formattingPortion = iter.next();
						}

						// add anonymous table
						blockBoxes.add(new TableBox(context, this, tableStartOffset, tableEndOffset));
						if (formattingPortion == null) {
							break;
						} else {
							iter.push(formattingPortion);
						}
					} else { // next is a block box element
						final INode blockNode = (INode) formattingPortion;
						blockBoxes.add(context.getBoxFactory().createBox(context, blockNode, this, width));
					}
				}
			}
		} else {
			final ContentRange range = new ContentRange(startOffset, endOffset);
			final InlineElementBox.InlineBoxes inlineBoxes = InlineElementBox.createInlineBoxes(context, node, range);
			pendingInlines.addAll(inlineBoxes.boxes);
			pendingInlines.add(new PlaceholderBox(context, node, range.getEndOffset() - node.getStartOffset()));
		}

		if (afterInlines != null) {
			pendingInlines.addAll(afterInlines);
		}

		if (!pendingInlines.isEmpty()) {
			blockBoxes.add(ParagraphBox.create(context, node, pendingInlines, width));
			pendingInlines.clear();
		}

		return blockBoxes;
	}

	protected boolean hasChildren() {
		return getChildren() != null && getChildren().length > 0;
	}

	/**
	 * Positions the children of this box. Vertical margins are collapsed here. Returns the vertical offset of the top
	 * of the first child to move, or -1 if not children were actually moved.
	 */
	protected int positionChildren(final LayoutContext context) {

		int childY = 0;
		int prevMargin = 0;
		final BlockBox[] children = getBlockChildren();
		int repaintStart = -1;

		Styles styles = null;

		if (!isAnonymous()) {
			styles = context.getStyleSheet().getStyles(getNode());
		}

		if (styles != null && children.length > 0) {
			if (styles.getBorderTopWidth() + styles.getPaddingTop().get(getWidth()) == 0) {
				// first child's top margin collapses into ours
				marginTop = Math.max(marginTop, children[0].getMarginTop());
				childY -= children[0].getMarginTop();
			}
		}

		for (int i = 0; i < children.length; i++) {

			final Insets insets = children[i].getInsets(context, getWidth());

			childY += insets.getTop();

			if (i > 0) {
				childY -= Math.min(prevMargin, children[i].getMarginTop());
			}

			if (repaintStart == -1 && children[i].getY() != childY) {
				repaintStart = Math.min(children[i].getY(), childY);
			}

			children[i].setX(insets.getLeft());
			children[i].setY(childY);

			childY += children[i].getHeight() + insets.getBottom();
			prevMargin = children[i].getMarginBottom();
		}

		if (styles != null && children.length > 0) {
			if (styles.getBorderBottomWidth() + styles.getPaddingBottom().get(getWidth()) == 0) {
				// last child's bottom margin collapses into ours
				marginBottom = Math.max(marginBottom, prevMargin);
				childY -= prevMargin;
			}
		}

		setHeight(childY);

		return repaintStart;
	}

	/**
	 * Sets the layout state of the box.
	 *
	 * @param layoutState
	 *            One of the LAYOUT_* constants
	 */
	protected void setLayoutState(final byte layoutState) {
		this.layoutState = layoutState;
	}

	// ========================================================= PRIVATE

	/**
	 * Return the end position of an anonymous box. The default implementation returns null.
	 */
	private IPosition getEndPosition() {
		return endPosition;
	}

	/**
	 * Return the start position of an anonymous box. The default implementation returns null.
	 */
	private IPosition getStartPosition() {
		return startPosition;
	}

	private boolean isTableChild(final LayoutContext context, final Object rangeOrElement) {
		if (rangeOrElement != null && rangeOrElement instanceof IElement) {
			return LayoutUtils.isTableChild(context.getStyleSheet(), (IElement) rangeOrElement);
		} else {
			return false;
		}
	}

}
